package org.eclipse.jdt.internal.compiler.tool;

import java.util.Iterator;
import java.util.ArrayList;
import java.nio.charset.CoderResult;
import java.nio.CharBuffer;
import javax.tools.FileObject;
import java.io.IOException;
import java.nio.charset.CharsetDecoder;
import java.nio.ByteBuffer;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.UnsupportedCharsetException;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.Charset;
import java.io.InputStream;

public final class Util
{
    public static String LINE_SEPARATOR;
    
    static {
        Util.LINE_SEPARATOR = System.getProperty("line.separator");
    }
    
    public Util() {
    }
    
    public static char[] getInputStreamAsCharArray(InputStream stream, int length, String encoding) throws IOException {
        Charset charset = null;
        try {
            charset = Charset.forName(encoding);
        }
        catch (IllegalCharsetNameException ex) {
            System.err.println("Illegal charset name : " + encoding);
            return null;
        }
        catch (UnsupportedCharsetException ex2) {
            System.err.println("Unsupported charset : " + encoding);
            return null;
        }
        CharsetDecoder charsetDecoder = charset.newDecoder();
        charsetDecoder.onMalformedInput(CodingErrorAction.REPLACE).onUnmappableCharacter(CodingErrorAction.REPLACE);
        byte[] contents = org.eclipse.jdt.internal.compiler.util.Util.getInputStreamAsByteArray(stream, length);
        ByteBuffer byteBuffer = ByteBuffer.allocate(contents.length);
        byteBuffer.put(contents);
        byteBuffer.flip();
        return charsetDecoder.decode(byteBuffer).array();
    }
    
    public static CharSequence getCharContents(FileObject fileObject, boolean ignoreEncodingErrors, byte[] contents, String encoding) throws IOException {
        if (contents == null) {
            return null;
        }
        Charset charset = null;
        try {
            charset = Charset.forName(encoding);
        }
        catch (IllegalCharsetNameException ex) {
            System.err.println("Illegal charset name : " + encoding);
            return null;
        }
        catch (UnsupportedCharsetException ex2) {
            System.err.println("Unsupported charset : " + encoding);
            return null;
        }
        CharsetDecoder charsetDecoder = charset.newDecoder();
        ByteBuffer byteBuffer = ByteBuffer.allocate(contents.length);
        byteBuffer.put(contents);
        byteBuffer.flip();
        if (ignoreEncodingErrors) {
            charsetDecoder.onMalformedInput(CodingErrorAction.REPLACE).onUnmappableCharacter(CodingErrorAction.REPLACE);
            return charsetDecoder.decode(byteBuffer);
        }
        charsetDecoder.onMalformedInput(CodingErrorAction.REPORT).onUnmappableCharacter(CodingErrorAction.REPORT);
        CharBuffer out = CharBuffer.allocate(contents.length);
        CoderResult result = null;
        String replacement = charsetDecoder.replacement();
        int replacementLength = replacement.length();
        Util.EncodingErrorCollector collector = null;
        while (true) {
            result = charsetDecoder.decode(byteBuffer, out, true);
            if (result.isMalformed() || result.isUnmappable()) {
                if (collector == null) {
                    collector = new Util.EncodingErrorCollector(fileObject, encoding);
                }
                reportEncodingError(collector, out.position(), result.length());
                if (out.position() + replacementLength >= out.capacity()) {
                    CharBuffer temp = CharBuffer.allocate(out.capacity() * 2);
                    out.flip();
                    temp.put(out);
                    out = temp;
                }
                out.append(replacement);
                byteBuffer.position(byteBuffer.position() + result.length());
            }
            else if (result.isOverflow()) {
                CharBuffer temp = CharBuffer.allocate(out.capacity() * 2);
                out.flip();
                temp.put(out);
                out = temp;
            }
            else {
                break;
            }
        }
        out.flip();
        if (collector != null) {
            collector.reportAllEncodingErrors(out.toString());
            return out;
        }
        return out;
    }
    
    private static void reportEncodingError(Util.EncodingErrorCollector collector, int position, int length) {
        collector.collect(position, -length);
    }
    
    public static class EncodingError
    {
        int position;
        int length;
        
        public EncodingError(int position, int length) {
            this.position = position;
            this.length = length;
        }
        
        public String getSource(char[] unitSource) {
            int startPosition = this.position;
            int endPosition = this.position + this.length - 1;
            if (startPosition > endPosition || (startPosition < 0 && endPosition < 0) || unitSource.length == 0) {
                return "No source available";
            }
            StringBuffer errorBuffer = new StringBuffer();
            errorBuffer.append('\t');
            int sourceLength = unitSource.length;
            int begin;
            char c;
            for (begin = ((startPosition >= sourceLength) ? (sourceLength - 1) : startPosition); begin > 0 && (c = unitSource[begin - 1]) != '\n' && c != '\r'; --begin) {}
            int end = (endPosition >= sourceLength) ? (sourceLength - 1) : endPosition;
            Label_0171: {
                while (end + 1 < sourceLength) {
                    if ((c = unitSource[end + 1]) != '\r' && c != '\n') {
                        ++end;
                        continue;
                    }
                    while ((c = unitSource[begin]) == ' ' || c == '\t') {
                        ++begin;
                    }
                    errorBuffer.append(unitSource, begin, end - begin + 1);
                    errorBuffer.append(Util.LINE_SEPARATOR).append("\t");
                    for (int i = begin; i < startPosition; ++i) {
                        errorBuffer.append((unitSource[i] == '\t') ? '\t' : ' ');
                    }
                    for (int i = startPosition; i <= ((endPosition >= sourceLength) ? (sourceLength - 1) : endPosition); ++i) {
                        errorBuffer.append('^');
                    }
                    return errorBuffer.toString();
                }
                continue;
            }
        }
    }
    
    public static class EncodingErrorCollector
    {
        ArrayList<Util.EncodingError> encodingErrors;
        FileObject fileObject;
        String encoding;
        
        public EncodingErrorCollector(FileObject fileObject, String encoding) {
            this.encodingErrors = new ArrayList();
            this.fileObject = fileObject;
            this.encoding = encoding;
        }
        
        public void collect(int position, int length) {
            this.encodingErrors.add(new EncodingError(position, length));
        }
        
        public void reportAllEncodingErrors(String string) {
            char[] unitSource = string.toCharArray();
            for (Util.EncodingError error : this.encodingErrors) {
                System.err.println(this.fileObject.getName() + " Unmappable character for encoding " + this.encoding);
                System.err.println(error.getSource(unitSource));
            }
        }
    }
}
